     /**
      * \brief JSON : Utilities
      *
      * Utilities module for the JSON service program.
      *
      * \author Mihael Schmidt
      * \date   31.01.2010
      *
      * \link http://www.rpgnextgen.com RPG Next Gen
      * \link http://www.json.org JSON
      *
      * \todo json_util_dataStructureToJsonObject : pointer to datastructure as parm
      *
      * \rev 01.08.2010 Mihael Schmidt
      *      added date data type support
      *
      * \rev 26.11.2010 Mihael Schmidt
      *      added toFile/fromFile procedures
      * 
      * \rev 09.04.2011 Mihael Schmidt
      *      added decimal data type support
      * 
      * \rev 06.05.2011 Mihael Schmidt
      *      Switch from Linked List to ArrayList
      * 
      * \rev 24.06.2011 Brian Garland
      *      added time and timestamp procedures,
      *      added dataStructure to JSON object procedure
      */

      *-------------------------------------------------------------------------
      * Copyright (c) 2010-2011 Mihael Schmidt
      * All rights reserved.
      *
      * This file is part of the JSON service program.
      *
      * JSON service program is free software: you can redistribute it and/or modify it under
      * the terms of the GNU Lesser General Public License as published by
      * the Free Software Foundation, either version 3 of the License, or
      * any later version.
      *
      * JSON service program is distributed in the hope that it will be useful,
      * but WITHOUT ANY WARRANTY; without even the implied warranty of
      * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
      * GNU Lesser General Public License for more details.
      *
      * You should have received a copy of the GNU Lesser General Public
      * License along with JSON service program. If not, see http://www.gnu.org/licenses/.
      *-------------------------------------------------------------------------


     H nomain
     H bnddir('QC2LE')
     H copyright('Copyright (c) 2010-2011 Mihael Schmidt. All rights reserved.')


      *-------------------------------------------------------------------------
      * Prototypes
      *-------------------------------------------------------------------------
     D extractNumber   PR           256A
     D  fieldPtr                       *   value
     D  editCode                      1A   value
     D  fieldType                    10A   value
     D  numDigits                     2P 0 value
     D  numDecimals                   2P 0 value
      * 
     D sendProgramMessage...
     D                 PR                  extpgm('QMHSNDPM')
     D  szMsgID                       7A   const
     D  szMsgFile                    20A   const
     D  szMsgData                  6000A   const  options(*varsize)
     D  nMsgDataLen                  10I 0 const
     D  szMsgType                    10A   const
     D  szCallStkEntry...
     D                               10A   const
     D  nRelativeCallStkEntry...
     D                               10I 0 const
     D  szRtnMsgKey                   4A
     D  error                       265A   options(*varsize)
      *
     D errno           PR              *   extproc('__errno')
      *
     D strerr          PR              *   extproc('strerror')
     D    errnum                     10I 0 value
      *
     D retreiveFileDescription...
     D                 PR                  extpgm('QDBRTVFD')
     D  receiverVariable...
     D                            65535A   options(*varsize)
     D  receiverLen                  10I 0 const
     D  qReturnFile                  20A
     D  formatName                    8A   const
     D  qFileName                    20A   const
     D  recordFormatName...
     D                               10A   const
     D  overrideProcessing...
     D                                1A   const
     D  system                       10A   const
     D  formatType                   10A   const
     D  errorCode                   256A   options(*varsize)
      *
     D convertEditCode...
     D                 PR                  extpgm('QECCVTEC')
     D  editMask                    256A
     D  editMaskLen                  10I 0
     D  outputLen                    10I 0
     D  zeroFillChar                  1A
     D  editCode                      1A   const
     D  fillChar                      1A   const
     D  numDigits                    10I 0 const
     D  numDecimals                  10I 0 const
     D  errorCode                   256A   options(*varsize)
      *
     D edit...
     D                 PR                  extpgm('QECEDT')
     D  editedField                 256A
     D  editedFieldLen...
     D                               10I 0
     D  field                        31A   const
     D  fieldType                    10A   const
     D  numDigits                    10I 0 const
     D  editMask                    256A   const
     D  editMaskLen                  10I 0 const
     D  zeroFillChar                  1A   const
     D  errorCode                   256A   options(*varsize)
      * 
      /copy 'json_h.rpgle'
      /copy 'json_int_h.rpgle'
      /copy 'ifsio_h.rpgle'
      /copy 'libc_h.rpgle'
      /copy 'ceeapi_h.rpgle'
      /copy 'arraylist_h.rpgle'

      /if not defined (QUSEC)
      /define QUSEC
      /copy QSYSINC/QRPGLESRC,QUSEC
      /endif


      *-------------------------------------------------------------------------
      * Constants
      *-------------------------------------------------------------------------
      * Field types returned from QDBRTVFD
     D FIELD_TYPE_BINARY...
     D                 C                   x'0000'
     D FIELD_TYPE_FLOAT...
     D                 C                   x'0001'
     D FIELD_TYPE_ZONED...
     D                 C                   x'0002'
     D FIELD_TYPE_PACKED...
     D                 C                   x'0003'
     D FIELD_TYPE_CHAR...
     D                 C                   x'0004'
     D FIELD_TYPE_VARCHAR...
     D                 C                   x'8004'
     D FIELD_TYPE_DATE...
     D                 C                   x'000B'
     D FIELD_TYPE_TIME...
     D                 C                   x'000C'
     D FIELD_TYPE_TIMESTAMP...
     D                 C                   x'000D'
      *
      /copy 'json_c.rpgle'


      *-------------------------------------------------------------------------
      * Global Variables
      *-------------------------------------------------------------------------
      *
      * Format Definition Header returned from QDBRTVFD
      *
     D Qddfmt          DS           256    BASED(p_Qddfmt)
     D  Qddbyrtn                     10I 0
     D  Qddbyava                     10I 0
     D  Qddftext                     50A   OVERLAY(Qddfmt:94)
     D  Qddffldnum                    5I 0 OVERLAY(Qddfmt:144)
      *
      * Field Header returned from QDBRTVFD
      *
     D Qddffld         DS           252    BASED(p_Qddffld)
     D  Qddfdefl                     10I 0
     D  Qddffldi                     30A
     D  Qddfflde                     30A
     D  Qddfftyp                      2A
     D  Qddffiob                      1A
     D  Qddffobo                     10I 0
     D  Qddffibo                     10I 0
     D  Qddffldb                      5I 0
     D  Qddffldd                      5I 0
     D  Qddffldp                      5I 0
     D  Qddffkbs                      1A
     D  Qddfxals                     10I 0 OVERLAY(Qddffld:237)
      *
      * Alias Name Structure returned from QDBRTVFD
      *
     D Qddfalis        DS                  BASED(p_Qddfalis)
     D  Qddfalsl                      5I 0
     D  Reserved_85                  14A
     D  Qddfalsn                    258A
      *
      * Parameters passed to CEEDOD
      *
     D descType        S             10I 0
     D dataType        S             10I 0
     D descInfo1       S             10I 0
     D descInfo2       S             10I 0
     D length          S             10I 0


      *-------------------------------------------------------------------------
      * Procedures
      *-------------------------------------------------------------------------

     /**
      * \brief Get date entry
      *
      * A convenience procedure for getting a date from a string entry.
      * The string must be in *ISO format (yyyy-mm-dd).
      *
      * <br><br>
      *
      * If the entry is not of type string or the entry is not a date string
      * in *ISO format an escape message will be sent.
      *
      * \author Mihael Schmidt
      * \date   01.08.2010
      *
      * \param Pointer to the JSON object
      * \param Key
      *
      * \return Date
      */
     P json_util_getDate...
     P                 B                   export
     D                 PI              D   opdesc
     D   json                          *   const
     D   key                      65535A   const options(*varsize)
      *
     D tmpKey          S          65535A
     D header          DS                  likeds(tmpl_header) based(json)
     D entry           DS                  likeds(tmpl_entry) based(entryPtr)

     D retVal          S               D
      /free
       isJSONImpl(json);

       cee_getOpDescInfo(2 : descType : dataType : descInfo1 : descInfo2 :
                         length : *omit);
       tmpKey = %subst(key : 1 : length);

       entryPtr = getJSONEntry(json : %addr(tmpKey) : length);
       if (entryPtr <> *null);
         if (entry.type = JSON_TYPE_STRING);
           retVal = %date(%str(entry.value) : *ISO);
         else;
           sendEscapeMessage(MSG_ENTRY_TYPE_MISMATCH);
         endif;
       endif;

       return retVal;
      /end-free
     P                 E


     /**
      * \brief Add date entry
      *
      * Convenience procedure for adding a date string in *ISO format to the
      * json object.
      *
      * \author Mihael Schmidt
      * \date 2010-08-01
      *
      * \param Pointer to the JSON object
      * \param Key
      * \param Date
      */
     P json_util_putDate...
     P                 B                   export
     D                 PI                  opdesc
     D   jsonPtr                       *   const
     D   key                      65535A   const options(*varsize)
     D   value                         D   const
      *
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
     D prevEntryPtr    S               *
     D prevEntry       DS                  likeds(tmpl_entry)
     D                                     based(prevEntryPtr)
     D newEntryPtr     S               *
     D newEntry        DS                  likeds(tmpl_entry)
     D                                     based(newEntryPtr)
     D tmpPtr          S               *
     D tmpKey          S          65535A
     D tmpValue        S             10A
     D keyLength       S             10I 0
     D hexnull         S              1A   inz(x'00')
      /free
       isJSONImpl(jsonPtr);

       cee_getOpDescInfo(2 : descType : dataType : descInfo1 : descInfo2 :
                         length : *omit);
       keyLength = length;
       tmpKey = %subst(key : 1 : keyLength);
       tmpValue = %char(value : *ISO);

       // check if key is already in the map
       newEntryPtr = getJSONEntry(jsonPtr : %addr(tmpKey) : keyLength);
       if (newEntryPtr = *null);    // entry not in json object
         // create new entry
         newEntryPtr = %alloc(%size(tmpl_entry));
         newEntry.type = JSON_TYPE_STRING;
         newEntry.keyLength = keyLength + 1;      // +1 for the null byte
         newEntry.key = %alloc(keyLength + 1);     // +1 for the null byte
         newEntry.valueLength = %size(tmpValue);
         newEntry.value = %alloc(%size(tmpValue) + 1);   // +1 for the null byte
         newEntry.next = *null;
         newEntry.prev = *null;

         // update header
         header.size += 1;
         if (header.firstEntry = *null);
           header.firstEntry = newEntryPtr;
         else;
           prevEntryPtr = header.lastEntry;
           prevEntry.next = newEntryPtr;
           newEntry.prev = prevEntryPtr;
         endif;
         header.lastEntry = newEntryPtr;

       else;
         // entry needs to be resized
         newEntry.type = JSON_TYPE_STRING;
         newEntry.keyLength = keyLength + 1;
         newEntry.key = %realloc(newEntry.key : keyLength + 1);
         newEntry.valueLength = %size(tmpValue);

         // dispose/dealloc old value
         disposeEntryValue(newEntryPtr);
       endif;

       // copy value to the list entry
       memcpy(newEntry.key : %addr(tmpKey) : keyLength);
       memcpy(newEntry.value : %addr(tmpValue) : %size(tmpValue));

       // set null to the last byte
       memcpy(newEntry.key + keyLength : %addr(hexNull) : 1);
       memcpy(newEntry.value + %size(tmpValue) : %addr(hexNull) : 1);
      /end-free
     P                 E


     /**
      * \brief Get date entry
      *
      * A convenience procedure for getting a date from a string entry.
      * The string must be in *ISO format (yyyy-mm-dd).
      *
      * <br><br>
      *
      * If the entry is not of type string or the entry is not a date string
      * in *ISO format an escape message will be sent.
      *
      * \author Mihael Schmidt
      * \date   01.08.2010
      *
      * \param Pointer to the JSON array
      * \param Index
      *
      * \return Date
      */
     P jsona_util_getDate...
     P                 B                   export
     D                 PI              D
     D   jsonArrPtr                    *   const
     D   index                       10U 0 const
      *
     D entry           DS                  likeds(tmpl_array_entry) based(ptr)
      /free
       ptr = arraylist_get(jsonArrPtr : index);
       if (ptr <> *null);
         if (entry.type = JSON_TYPE_STRING);
           return %date(%str(ptr + 4) : *ISO);
         else;
           sendEscapeMessage(MSG_ENTRY_TYPE_MISMATCH);
         endif;
       else;
         sendEscapeMessage(MSG_ARRAY_ELEMENT_NOT_FOUND);
       endif;

       return %date(); // dummy => will never come to this point
      /end-free
     P                 E


     /**
      * \brief Add date entry
      *
      * Convenience procedure for adding a date string in *ISO format to the
      * JSON array.
      *
      * \author Mihael Schmidt
      * \date   01.08.2010
      *
      * \param Pointer to the JSON array
      * \param Date
      * \param Index (default: last)
      */
     P jsona_util_putDate...
     P                 B                   export
     D                 PI
     D   jsonArrPtr                    *   const
     D   value                         D   const
     D   index                       10U 0 const options(*nopass)
      *
     D entry           DS                  likeds(tmpl_array_entry) inz
      /free
       entry.type = JSON_TYPE_STRING;
       entry.string = %char(value : *ISO);

       if (%parms() = 2);
         arraylist_add(jsonArrPtr : %addr(entry) : 10 + 4);
       else;
         arraylist_add(jsonArrPtr : %addr(entry) : 10 + 4 : index);
       endif;
      /end-free
     P                 E



     /**
      * \brief Output JSON Object to stream file
      *
      * \param JSON Object
      * \param Stream file path
      * \param CCSID (optional)
      * \param Delete if file exist (optional)
      *
      * \throws CPF9898 Error on opening file
      */
     P json_util_toFile...
     P                 B                   export
     D                 PI
     D  json                           *   const
     D  filePath                   1024A   const
     D  pCcsid                       10I 0 const options(*nopass : *omit)
     D  pDeleteFile                    N   const options(*nopass : *omit)
     D  pOutputFormat...
     D                               10I 0 const options(*nopass)
     D  pPrettyPrintFormat...
     D                                     likeds(json_prettyPrintFormat)
     D                                     const options(*nopass)
      *
     D outputFormat    S             10I 0
     D prettyPrintOption...
     D                 S               N   inz(*off)
     D prettyPrintFormat...
     D                 DS                  likeds(json_prettyPrintFormat) inz
     D ccsid           S             10I 0 inz(0)
     D deleteFile      S               N   inz(*off)
     D jsonString      S               *
     D fileHandle      S             10I 0
     D flags           S             10I 0
     D mode            S             10I 0
     D errorNumber     S             10I 0 based(errorPtr)
     D errorString     S           1024A   varying
     D msgdata         S            512A
     D msgkey          S              4A

      /free
       if (%parms() >= 3 and %addr(pCcsid) <> *null);
         ccsid = pCcsid;
       endif;

       if (%parms() >= 4 and %addr(pDeleteFile) <> *null);
         deleteFile = pDeleteFile;
       endif;

       if (%parms() >= 5);
         outputFormat = pOutputFormat;
         prettyPrintFormat.useTabs = *off;
         prettyPrintFormat.numberWhitespace = 4;
         prettyPrintFormat.breakAfterElement = *on;
         prettyPrintFormat.breakAfterObjectBrace = *on;
         prettyPrintFormat.breakBeforeArrayBrace = *on;
         prettyPrintFormat.breakAfterArrayBrace = *on;
       endif;

       if (%parms() >= 6);
         prettyPrintFormat = pPrettyPrintFormat;
       endif;

       if (outputFormat = JSON_OUTPUT_PRETTY_PRINT);
         prettyPrintOption = *on;
       endif;


       if (deleteFile);
         if (access(%trimr(filePath) : F_OK) >= 0);
           // file exists => delete
           system('RMVLNK OBJLNK(''' + %trimr(filePath) + ''')');
         endif;
       endif;


       // create file
       flags = O_WRONLY + O_CREAT + O_APPEND + O_TEXTDATA +
               O_CCSID + O_TEXT_CREAT;
       mode =  S_IRUSR + S_IWUSR + S_IRGRP + S_IROTH;
       fileHandle = open(%trimr(filePath) : flags : mode : ccsid : 0);
       if (fileHandle < 0);
         // could not create file
         errorPtr = errno();
         errorString = 'I/O Error: ' + %str(strerr(errorNumber));
         msgdata = 'Could not create file ' + %trimr(filePath) +
                   ': ' + errorString;
         clear QUSEC;
         QUSBPRV = 0;
         sendProgramMessage('CPF9898' :
                            'QCPFMSG   *LIBL     ' :
                            %trimr(msgdata) :
                            %len(%trimr(msgdata)) :
                            '*ESCAPE   ' :
                            '*PGMBDY' :
                            0 :
                            msgkey :
                            QUSEC);
       endif;

       jsonString = json_toString(json  : outputFormat : prettyPrintFormat);

       callp write(fileHandle : jsonString : strlen(jsonString));

       dealloc(n) jsonString;

       callp close(fileHandle);
      /end-free
     P                 E


     /**
      * \brief Read JSON from stream file
      *
      * \param Stream file path
      * \param Delete file after reading (optional)
      *
      * \return JSON object
      *
      * \throws CPF9898 Error on reading file
      * \throws CPF9898 File not found
      * \throws CPF9898 Error buildng JSON object
      */
     P json_util_fromFile...
     P                 B                   export
     D                 PI              *
     D  filePath                   1024A   const
     D  pDeleteFile                    N   const options(*nopass)
      *
     D json            S               *
     D deleteFile      S               N   inz(*off)
     D buffer          S               *
     D fileSize        S             10I 0
     D fileStatus      DS                  likeds(statds)
     D fileHandle      S             10I 0
     D flags           S             10I 0
     D errorNumber     S             10I 0 based(errorPtr)
     D errorString     S           1024A   varying
     D msgdata         S            512A
     D msgkey          S              4A

      /free
       if (%parms() = 2);
         deleteFile = pDeleteFile;
       endif;

       if (stat(%trimr(filePath) : %addr(fileStatus)) < 0);
         msgdata = 'Could not query file attributes for file ' +
                   %trimr(filepath) + '.';
         clear QUSEC;
         QUSBPRV = 0;
         sendProgramMessage('CPF9898' :
                            'QCPFMSG   *LIBL     ' :
                            %trimr(msgdata) :
                            %len(%trimr(msgdata)) :
                            '*ESCAPE   ' :
                            '*PGMBDY' :
                            0 :
                            msgkey :
                            QUSEC);
       endif;

       flags = O_RDONLY + O_TEXTDATA;
       fileHandle = open(%trimr(filePath) : flags);
       if (fileHandle < 0);
         // could not open file
         errorPtr = errno();
         errorString = 'I/O Error: ' + %str(strerr(errorNumber));
         msgdata = 'Could not open file ' + %trimr(filePath) + ': '+errorString;
         clear QUSEC;
         QUSBPRV = 0;
         sendProgramMessage('CPF9898' :
                            'QCPFMSG   *LIBL     ' :
                            %trimr(msgdata) :
                            %len(%trimr(msgdata)) :
                            '*ESCAPE   ' :
                            '*PGMBDY' :
                            0 :
                            msgkey :
                            QUSEC);
       endif;

       buffer = %alloc(fileStatus.st_size);

       callp read(fileHandle : buffer : fileStatus.st_size);

       callp close(fileHandle);

       json = json_parse(buffer);

       // clean up
       dealloc buffer;

       if (deleteFile);
         if (access(%trimr(filePath) : F_OK) >= 0);
           // file exists => delete
           system('RMVLNK OBJLNK(''' + %trimr(filePath) + ''')');
         endif;
       endif;

       return json;
      /end-free
     P                 E


     /**
      * \brief Get decimal entry
      *
      * A convenience procedure for getting a decimal value from a string entry.
      *
      * <br><br>
      *
      * If the entry is not of type string or the entry is not a number
      * an escape message will be sent.
      *
      * \author Mihael Schmidt
      * \date   09.04.2011
      *
      * \param Pointer to the JSON object
      * \param Key
      *
      * \return Decimal value
      */
     P json_util_getDecimal...
     P                 B                   export
     D                 PI            30P10 opdesc
     D   json                          *   const
     D   key                      65535A   const options(*varsize)
      *
     D tmpKey          S          65535A
     D header          DS                  likeds(tmpl_header) based(json)
     D entry           DS                  likeds(tmpl_entry) based(entryPtr)

     D retVal          S             30P10
      /free
       isJSONImpl(json);

       cee_getOpDescInfo(2 : descType : dataType : descInfo1 : descInfo2 :
                         length : *omit);
       tmpKey = %subst(key : 1 : length);

       entryPtr = getJSONEntry(json : %addr(tmpKey) : length);
       if (entryPtr <> *null);
         if (entry.type = JSON_TYPE_STRING);
           retVal = %dec(%str(entry.value) : 30 : 10);
         else;
           sendEscapeMessage(MSG_ENTRY_TYPE_MISMATCH);
         endif;
       endif;

       return retVal;
      /end-free
     P                 E
    
    
     /**
      * \brief Get decimal entry
      *
      * A convenience procedure for getting a decimal value from a string entry.
      *
      * <br><br>
      *
      * If the entry is not of type string or the entry is not a number
      * an escape message will be sent.
      *
      * \author Mihael Schmidt
      * \date   09.04.2011
      *
      * \param Pointer to the JSON array 
      * \param Index
      *
      * \return Decimal value
      */
     P jsona_util_getDecimal...
     P                 B                   export
     D                 PI            30P10
     D   jsonArrPtr                    *   const
     D   index                       10U 0 const
      *
     D entry           DS                  likeds(tmpl_array_entry) based(ptr)
     D retVal          S             30P10
      /free
       ptr = arraylist_get(jsonArrPtr : index);
       if (ptr <> *null);
         if (entry.type = JSON_TYPE_STRING);
           retVal = %dec(%str(ptr + 4) : 30 : 10);
           return retVal;
         else;
           sendEscapeMessage(MSG_ENTRY_TYPE_MISMATCH);
         endif;
       else;
         sendEscapeMessage(MSG_ARRAY_ELEMENT_NOT_FOUND);
       endif;

       return 0; // dummy => will never come to this point
      /end-free
     P                 E
     
     
     /**
      * \brief Add decimal entry
      *
      * Convenience procedure for adding a decimal value as a string to the 
      * json object.
      *
      * \author Mihael Schmidt
      * \date   09.04.2011
      *
      * \param Pointer to the JSON object
      * \param Key
      * \param Decimal value
      * \param Decimal positions (default: 2)
      */
     P json_util_putDecimal...
     P                 B                   export
     D                 PI                  opdesc
     D   jsonPtr                       *   const
     D   key                      65535A   const options(*varsize)
     D   value                       30P10 const
     D   pDecimalPositions...
     D                                2P 0 const options(*nopass)
      *
     D decimalPositions...
     D                 S              2P 0 inz(2)
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
     D prevEntryPtr    S               *
     D prevEntry       DS                  likeds(tmpl_entry)
     D                                     based(prevEntryPtr)
     D newEntryPtr     S               *
     D newEntry        DS                  likeds(tmpl_entry)
     D                                     based(newEntryPtr)
     D tmpPtr          S               *
     D tmpKey          S          65535A
     D tmpValue        S             32A
     D tmpValueLength  S             10I 0
     D keyLength       S             10I 0
     D hexnull         S              1A   inz(x'00')
     D decimalSepIndex...
     D                 S             10I 0
      /free
       isJSONImpl(jsonPtr);

       if (%parms() = 4);
         decimalPositions = pDecimalPositions;
       endif;

       cee_getOpDescInfo(2 : descType : dataType : descInfo1 : descInfo2 :
                         length : *omit);
       keyLength = length;
       tmpKey = %subst(key : 1 : keyLength);
       tmpValue = %char(value);

       // max. decimal precision of 10
       if (decimalPositions > 10);
         decimalPositions = 10;
       elseif (decimalPositions < 0); // min. decimal precision of 0
         decimalPositions = 0;
       endif;

       // limit the string to the passed number of decimal positions
       decimalSepIndex = %scan('.' : tmpValue);
       if (decimalSepIndex = 0);
         decimalSepIndex = %scan(',' : tmpValue);
       endif;
       
       if (decimalSepIndex <> 0);
         tmpValueLength = decimalSepIndex + decimalPositions; 
       else;
         tmpValueLength = %len(%trim(tmpvalue));
       endif;

       // check if key is already in the map
       newEntryPtr = getJSONEntry(jsonPtr : %addr(tmpKey) : keyLength);
       if (newEntryPtr = *null);    // entry not in json object
         // create new entry
         newEntryPtr = %alloc(%size(tmpl_entry));
         newEntry.type = JSON_TYPE_STRING;
         newEntry.keyLength = keyLength + 1;      // +1 for the null byte
         newEntry.key = %alloc(keyLength + 1);     // +1 for the null byte
         newEntry.valueLength = tmpValueLength;
         newEntry.value = %alloc(tmpValueLength + 1);   // +1 for the null byte
         newEntry.next = *null;
         newEntry.prev = *null;

         // update header
         header.size += 1;
         if (header.firstEntry = *null);
           header.firstEntry = newEntryPtr;
         else;
           prevEntryPtr = header.lastEntry;
           prevEntry.next = newEntryPtr;
           newEntry.prev = prevEntryPtr;
         endif;
         header.lastEntry = newEntryPtr;

       else;
         // entry needs to be resized
         newEntry.type = JSON_TYPE_STRING;
         newEntry.keyLength = keyLength + 1;
         newEntry.key = %realloc(newEntry.key : keyLength + 1);
         newEntry.valueLength = tmpValueLength;

         // dispose/dealloc old value
         disposeEntryValue(newEntryPtr);
       endif;

       // copy value to the list entry
       memcpy(newEntry.key : %addr(tmpKey) : keyLength);
       memcpy(newEntry.value : %addr(tmpValue) : tmpValueLength);

       // set null to the last byte
       memcpy(newEntry.key + keyLength : %addr(hexNull) : 1);
       memcpy(newEntry.value + tmpValueLength : %addr(hexNull) : 1);
      /end-free
     P                 E
    
    
     /**
      * \brief Add decimal entry
      *
      * Convenience procedure for adding a decimal value as a string to the
      * JSON array.
      *
      * \author Mihael Schmidt
      * \date   12.04.2011
      *
      * \param Pointer to the JSON array
      * \param Decimal value
      * \param Decimal positions (default: 2)
      * \param Index (default: last)
      */
     P jsona_util_putDecimal...
     P                 B                   export
     D                 PI
     D   jsonArrPtr                    *   const
     D   value                       30P10 const
     D   pDecimalPositions...
     D                                2P 0 const options(*nopass : *omit)
     D   index                       10U 0 const options(*nopass)
      *
     D entry           DS                  likeds(tmpl_array_entry) inz
     D decimalSepIndex...
     D                 S             10I 0
     D decimalPositions...
     D                 S              2P 0 inz(2)
     D tmpValueLength  S             10I 0
      /free
       entry.type = JSON_TYPE_STRING;
       entry.string = %char(value);
       
       if (%parms() >= 3 and %addr(pDecimalPositions) <> *null);
         decimalPositions = pDecimalPositions;
       endif;

       // max. decimal precision of 10
       if (decimalPositions > 10);
         decimalPositions = 10;
       elseif (decimalPositions < 0); // min. decimal precision of 0
         decimalPositions = 0;
       endif;
       
       // limit the string to the passed number of decimal positions
       decimalSepIndex = %scan('.' : entry.string);
       if (decimalSepIndex = 0);
         decimalSepIndex = %scan(',' : entry.string);
       endif;
       
       if (decimalSepIndex <> 0);
         tmpValueLength = decimalSepIndex + decimalPositions; 
       else;
         tmpValueLength = %len(%trim(entry.string));
       endif;

       if (%parms() < 4);
         arraylist_add(jsonArrPtr : %addr(entry) : tmpValueLength + 4);
       else;
         arraylist_add(jsonArrPtr : %addr(entry) : tmpValueLength + 4 : index);
       endif;
      /end-free
     P                 E
    
    
    
     /**
      * \brief Get time entry
      *
      * A convenience procedure for get a time from a string entry.
      * The string must be in *ISO format (hh.mm.ss).
      *
      * <br><br>
      *
      * If the entry is not of type string or the entry is not a time string
      * in *ISO format an escape message will be send.
      *
      * \author Brian Garland
      * \date   2011-06-24
      *
      * \param Pointer to the JSON object
      * \param Key
      *
      * \return Time
      */
     P json_util_getTime...
     P                 B                   export
     D                 PI              T   opdesc
     D   json                          *   const
     D   key                      65535A   const options(*varsize)
      *
     D tmpKey          S          65535A
     D header          DS                  likeds(tmpl_header) based(json)
     D entry           DS                  likeds(tmpl_entry) based(entryPtr)

     D retVal          S               T
      /free
       isJSONImpl(json);

       cee_getOpDescInfo(2 : descType : dataType : descInfo1 : descInfo2 :
                         length : *omit);
       tmpKey = %subst(key : 1 : length);

       entryPtr = getJSONEntry(json : %addr(tmpKey) : length);
       if (entryPtr <> *null);
         if (entry.type = JSON_TYPE_STRING);
           retVal = %time(%str(entry.value) : *ISO);
         else;
           sendEscapeMessage(MSG_ENTRY_TYPE_MISMATCH);
         endif;
       endif;

       return retVal;
      /end-free
     P                 E



     /**
      * \brief Add time entry
      *
      * Convenience procedure for adding a time string in *ISO format to the
      * json object.
      *
      * \author Brian Garland
      * \date   2011-06-24
      *
      * \param Pointer to the JSON object
      * \param Key
      * \param Time
      */
     P json_util_putTime...
     P                 B                   export
     D                 PI                  opdesc
     D   jsonPtr                       *   const
     D   key                      65535A   const options(*varsize)
     D   value                         T   const
      *
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
     D prevEntryPtr    S               *
     D prevEntry       DS                  likeds(tmpl_entry)
     D                                     based(prevEntryPtr)
     D newEntryPtr     S               *
     D newEntry        DS                  likeds(tmpl_entry)
     D                                     based(newEntryPtr)
     D tmpPtr          S               *
     D tmpKey          S          65535A
     D tmpValue        S              8A
     D keyLength       S             10I 0
     D hexnull         S              1A   inz(x'00')
      /free
       isJSONImpl(jsonPtr);

       cee_getOpDescInfo(2 : descType : dataType : descInfo1 : descInfo2 :
                         length : *omit);
       keyLength = length;
       tmpKey = %subst(key : 1 : keyLength);
       tmpValue = %char(value : *ISO);

       // check if key is already in the map
       newEntryPtr = getJSONEntry(jsonPtr : %addr(tmpKey) : keyLength);
       if (newEntryPtr = *null);    // entry not in json object
         // create new entry
         newEntryPtr = %alloc(%size(tmpl_entry));
         newEntry.type = JSON_TYPE_STRING;
         newEntry.keyLength = keyLength + 1;      // +1 for the null byte
         newEntry.key = %alloc(keyLength + 1);     // +1 for the null byte
         newEntry.valueLength = %size(tmpValue);
         newEntry.value = %alloc(%size(tmpValue) + 1);   // +1 for the null byte
         newEntry.next = *null;
         newEntry.prev = *null;

         // update header
         header.size += 1;
         if (header.firstEntry = *null);
           header.firstEntry = newEntryPtr;
         else;
           prevEntryPtr = header.lastEntry;
           prevEntry.next = newEntryPtr;
           newEntry.prev = prevEntryPtr;
         endif;
         header.lastEntry = newEntryPtr;

       else;
         // entry needs to be resized
         newEntry.type = JSON_TYPE_STRING;
         newEntry.keyLength = keyLength + 1;
         newEntry.key = %realloc(newEntry.key : keyLength + 1);
         newEntry.valueLength = %size(tmpValue);

         // dispose/dealloc old value
         disposeEntryValue(newEntryPtr);
       endif;

       // copy value to the list entry
       memcpy(newEntry.key : %addr(tmpKey) : keyLength);
       memcpy(newEntry.value : %addr(tmpValue) : %size(tmpValue));

       // set null to the last byte
       memcpy(newEntry.key + keyLength : %addr(hexNull) : 1);
       memcpy(newEntry.value + %size(tmpValue) : %addr(hexNull) : 1);
      /end-free
     P                 E


     /**
      * \brief Get time entry
      *
      * A convenience procedure for get a time from a string entry.
      * The string must be in *ISO format (hh.mm.ss).
      *
      * <br><br>
      *
      * If the entry is not of type string or the entry is not a time string
      * in *ISO format an escape message will be send.
      *
      * \author Brian Garland
      * \date   2011-06-24
      *
      * \param Pointer to the JSON array
      * \param Index
      *
      * \return Time
      */
     P jsona_util_getTime...
     P                 B                   export
     D                 PI              T
     D   jsonArrPtr                    *   const
     D   index                       10U 0 const
      *
     D entry           DS                  likeds(tmpl_array_entry) based(ptr)
      /free
       ptr = arraylist_get(jsonArrPtr : index);
       if (ptr <> *null);
         if (entry.type = JSON_TYPE_STRING);
           return %time(%str(ptr + 4) : *ISO);
         else;
           sendEscapeMessage(MSG_ENTRY_TYPE_MISMATCH);
         endif;
       else;
         sendEscapeMessage(MSG_ARRAY_ELEMENT_NOT_FOUND);
       endif;

       return %time(); // dummy => will never come to this point
      /end-free
     P                 E


     /**
      * \brief Add time entry
      *
      * Convenience procedure for adding a time string in *ISO format to the
      * JSON array.
      *
      * \author Brian Garland
      * \date   2011-06-24
      *
      * \param Pointer to the JSON array
      * \param Time
      * \param Index (default: last)
      */
     P jsona_util_putTime...
     P                 B                   export
     D                 PI
     D   jsonArrPtr                    *   const
     D   value                         T   const
     D   index                       10U 0 const options(*nopass)
      *
     D entry           DS                  likeds(tmpl_array_entry) inz
      /free
       entry.type = JSON_TYPE_STRING;
       entry.string = %char(value : *ISO);

       if (%parms() = 2);
         arraylist_add(jsonArrPtr : %addr(entry) : 8 + 4);
       else;
         arraylist_add(jsonArrPtr : %addr(entry) : 8 + 4 : index);
       endif;
      /end-free
     P                 E



     /**
      * \brief Get timestamp entry
      *
      * A convenience procedure for get a timestamp from a string entry.
      * The string must be in *ISO format (yyyy-mm-dd-hh.mm.ss.mmmmmm).
      *
      * <br><br>
      *
      * If the entry is not of type string or the entry is not a timestamp string
      * in *ISO format an escape message will be send.
      *
      * \author Brian Garland
      * \date   2011-06-24
      *
      * \param Pointer to the JSON object
      * \param Key
      *
      * \return Timestamp
      */
     P json_util_getTimestamp...
     P                 B                   export
     D                 PI              Z   opdesc
     D   json                          *   const
     D   key                      65535A   const options(*varsize)
      *
     D tmpKey          S          65535A
     D header          DS                  likeds(tmpl_header) based(json)
     D entry           DS                  likeds(tmpl_entry) based(entryPtr)

     D retVal          S               Z
      /free
       isJSONImpl(json);

       cee_getOpDescInfo(2 : descType : dataType : descInfo1 : descInfo2 :
                         length : *omit);
       tmpKey = %subst(key : 1 : length);

       entryPtr = getJSONEntry(json : %addr(tmpKey) : length);
       if (entryPtr <> *null);
         if (entry.type = JSON_TYPE_STRING);
           retVal = %timestamp(%str(entry.value) : *ISO);
         else;
           sendEscapeMessage(MSG_ENTRY_TYPE_MISMATCH);
         endif;
       endif;

       return retVal;
      /end-free
     P                 E


     /**
      * \brief Add timestamp entry
      *
      * Convenience procedure for adding a timestamp string in *ISO format to the
      * json object.
      *
      * \author Brian Garland
      * \date   2011-06-24
      *
      * \param Pointer to the JSON object
      * \param Key
      * \param Timestamp
      */
     P json_util_putTimestamp...
     P                 B                   export
     D                 PI                  opdesc
     D   jsonPtr                       *   const
     D   key                      65535A   const options(*varsize)
     D   value                         Z   const
      *
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
     D prevEntryPtr    S               *
     D prevEntry       DS                  likeds(tmpl_entry)
     D                                     based(prevEntryPtr)
     D newEntryPtr     S               *
     D newEntry        DS                  likeds(tmpl_entry)
     D                                     based(newEntryPtr)
     D tmpPtr          S               *
     D tmpKey          S          65535A
     D tmpValue        S             26A
     D keyLength       S             10I 0
     D hexnull         S              1A   inz(x'00')
      /free
       isJSONImpl(jsonPtr);

       cee_getOpDescInfo(2 : descType : dataType : descInfo1 : descInfo2 :
                         length : *omit);
       keyLength = length;
       tmpKey = %subst(key : 1 : keyLength);
       tmpValue = %char(value : *ISO);

       // check if key is already in the map
       newEntryPtr = getJSONEntry(jsonPtr : %addr(tmpKey) : keyLength);
       if (newEntryPtr = *null);    // entry not in json object
         // create new entry
         newEntryPtr = %alloc(%size(tmpl_entry));
         newEntry.type = JSON_TYPE_STRING;
         newEntry.keyLength = keyLength + 1;      // +1 for the null byte
         newEntry.key = %alloc(keyLength + 1);     // +1 for the null byte
         newEntry.valueLength = %size(tmpValue);
         newEntry.value = %alloc(%size(tmpValue) + 1);   // +1 for the null byte
         newEntry.next = *null;
         newEntry.prev = *null;

         // update header
         header.size += 1;
         if (header.firstEntry = *null);
           header.firstEntry = newEntryPtr;
         else;
           prevEntryPtr = header.lastEntry;
           prevEntry.next = newEntryPtr;
           newEntry.prev = prevEntryPtr;
         endif;
         header.lastEntry = newEntryPtr;

       else;
         // entry needs to be resized
         newEntry.type = JSON_TYPE_STRING;
         newEntry.keyLength = keyLength + 1;
         newEntry.key = %realloc(newEntry.key : keyLength + 1);
         newEntry.valueLength = %size(tmpValue);

         // dispose/dealloc old value
         disposeEntryValue(newEntryPtr);
       endif;

       // copy value to the list entry
       memcpy(newEntry.key : %addr(tmpKey) : keyLength);
       memcpy(newEntry.value : %addr(tmpValue) : %size(tmpValue));

       // set null to the last byte
       memcpy(newEntry.key + keyLength : %addr(hexNull) : 1);
       memcpy(newEntry.value + %size(tmpValue) : %addr(hexNull) : 1);
      /end-free
     P                 E


     /**
      * \brief Get timestamp entry
      *
      * A convenience procedure for get a timestamp from a string entry.
      * The string must be in *ISO format (yyyy-mm-dd-hh.mm.ss.mmmmmm).
      *
      * <br><br>
      *
      * If the entry is not of type string or the entry is not a timestamp string
      * in *ISO format an escape message will be send.
      *
      * \author Brian Garland
      * \date   2011-06-24
      *
      * \param Pointer to the JSON array
      * \param Index
      *
      * \return Timestamp
      */
     P jsona_util_getTimestamp...
     P                 B                   export
     D                 PI              Z
     D   jsonArrPtr                    *   const
     D   index                       10U 0 const
      *
     D entry           DS                  likeds(tmpl_array_entry) based(ptr)
      /free
       ptr = arraylist_get(jsonArrPtr : index);
       if (ptr <> *null);
         if (entry.type = JSON_TYPE_STRING);
           return %timestamp(%str(ptr + 4) : *ISO);
         else;
           sendEscapeMessage(MSG_ENTRY_TYPE_MISMATCH);
         endif;
       else;
         sendEscapeMessage(MSG_ARRAY_ELEMENT_NOT_FOUND);
       endif;

       return %timestamp(); // dummy => will never come to this point
      /end-free
     P                 E


     /**
      * \brief Add timestamp entry
      *
      * Convenience procedure for adding a timestamp string in *ISO format to the
      * JSON array.
      *
      * \author Brian Garland
      * \date   2011-06-24
      *
      * \param Pointer to the JSON array
      * \param Timestamp
      * \param Index (default: last)
      */
     P jsona_util_putTimestamp...
     P                 B                   export
     D                 PI
     D   jsonArrPtr                    *   const
     D   value                         Z   const
     D   index                       10U 0 const options(*nopass)
      *
     D entry           DS                  likeds(tmpl_array_entry) inz
      /free
       entry.type = JSON_TYPE_STRING;
       entry.string = %char(value : *ISO);

       if (%parms() = 2);
         arraylist_add(jsonArrPtr : %addr(entry) : 26 + 4);
       else;
         arraylist_add(jsonArrPtr : %addr(entry) : 26 + 4 : index);
       endif;
      /end-free
     P                 E


     /**
      * \brief Convert a data structure to a JSON object
      *
      * \param Data structure
      * \param Qualified file name that the data structure is based upon
      *        (10A filename + 10A library)
      *
      * \return JSON object
      *
      * \throws CPF9898 File not found
      * \throws CPF9898 Error building JSON object
      * 
      * \info Varying fields with a 4-byte header are not supported.
      * 
      * \author Brian Garland
      */
     P json_util_dsToJsonObject...
     P                 B                   export
     D                 PI              *
     D   dataStructure...
     D                            65535A   options(*varsize)
     D   qFile                       20A   const
      *
      * Used to overlay some field types that either the API's don't support
      * or it is just more convenient.
      *
     D fieldDS         DS                  based(p_FieldDS) qualified
     D  integer                       9B 0 overlay(fieldDS:1)
     D  smallInt                      4B 0 overlay(fieldDS:1)
     D  bigInt                       20I 0 overlay(fieldDS:1)
     D  float                         4F   overlay(fieldDS:1)
     D  date                         10D   overlay(fieldDS:1)
     D  time                          8T   overlay(fieldDS:1)
     D  timestamp                    26Z   overlay(fieldDS:1)
     D  real                          4F   overlay(fieldDS:1)
     D  double                        8F   overlay(fieldDS:1)
      *
     D buffer          S              1A   based(p_Buffer)
     D bufferSize      S             10I 0
     D data            S             30A
     D fieldName       S             50A
     D i               S             10I 0
     D json            S               *
     D msgdata         S            512A
     D msgkey          S              4A
     D nbrInt          S             10I 0
     D nbrDouble       S              8F
     D nbrLong         S             20I 0
     D rtnFile         S             20A
     D varCharLen      S              5I 0 BASED(p_varCharLen)

      /free

       // Allocate just enough memory to find out how much we really need
       p_Buffer = %ALLOC(8);
       clear QUSEC;
       QUSBPRV = %SIZE(QUSEC);
       RetreiveFileDescription(buffer :
                               8 :
                               rtnFile :
                               'FILD0200' :
                               qFile :
                               '*FIRST' :
                               '0' :
                               '*LCL' :
                               '*EXT' :
                               QUSEC);
       if QUSBAVL > 0;
         // could not open file
         msgdata = 'Error retreiving file description for ' +
                   %trimr(%subst(qFile:11:10)) + '/' +
                   %trimr(%subst(qFile:1:10)) + '.  (' + QUSEI + ')';
         clear QUSEC;
         QUSBPRV = 0;
         sendProgramMessage('CPF9898' :
                            'QCPFMSG   *LIBL     ' :
                            %trimr(msgdata) :
                            %len(%trimr(msgdata)) :
                            '*ESCAPE   ' :
                            '*PGMBDY' :
                            0 :
                            msgkey :
                            QUSEC);
       endif;

       // The API tells us how much memory we really need
       p_Qddfmt = p_Buffer;
       bufferSize = Qddbyava;

       // Now allocate the correct amount of memory and call again
       p_Buffer = %REALLOC(p_Buffer:bufferSize);
       clear QUSEC;
       QUSBPRV = %SIZE(QUSEC);
       RetreiveFileDescription(buffer :
                               bufferSize :
                               rtnFile :
                               'FILD0200' :
                               qFile :
                               '*FIRST' :
                               '0' :
                               '*LCL' :
                               '*EXT' :
                               QUSEC);
       if QUSBAVL > 0;
         // could not open file
         msgdata = 'Error retreiving file description for ' +
                   %trimr(%subst(qFile:11:10)) + '/' +
                   %trimr(%subst(qFile:1:10)) + '.  (' + QUSEI + ')';
         clear QUSEC;
         QUSBPRV = 0;
         sendProgramMessage('CPF9898' :
                            'QCPFMSG   *LIBL     ' :
                            %trimr(msgdata) :
                            %len(%trimr(msgdata)) :
                            '*ESCAPE   ' :
                            '*PGMBDY' :
                            0 :
                            msgkey :
                            QUSEC);
       endif;

       json = json_Create();

       // Point to the Format Definition Header
       p_Qddfmt = p_Buffer;

       // Point to the first Field Header
       p_Qddffld = p_Qddfmt + %size(Qddfmt);

       for i = 1 to Qddffldnum;

         // Point to Alias Name Entry
         p_Qddfalis = p_Qddffld + Qddfxals;

         // Determine field name to use as JSON "name"
         // Qddffldi = field name                      aka: short name
         // Qddfalsn = alias name (length = Qddfalsl)  aka: long name
         if Qddfalsl = 0;
             fieldName = Qddffldi;
         else;
             fieldName = %subst(Qddfalsn:1:Qddfalsl);
         endif;

         // Retreive actual data to use as JSON "value"
         // Qddfftyp = field type (see constants)
         // Qddffibo = offset to field
         // Qddffldb = field length
         // Qddffldd = number of digits (for numeric)
         // Qddffldp = decimal positions (for numeric)
         select;
           when Qddfftyp = FIELD_TYPE_BINARY; // either 5I0, 10I0, or 20I0
             p_FieldDS = %ADDR(dataStructure)+Qddffibo;
             if Qddffldb = 2;
               json_putInt(json:%TRIMR(fieldName):fieldDS.smallInt);
             elseif Qddffldb = 4;
               json_putInt(json:%TRIMR(fieldName):fieldDS.integer);
             elseif Qddffldb = 8;
               json_putLong(json:%TRIMR(fieldName):fieldDS.bigInt);
             endif;

           when Qddfftyp = FIELD_TYPE_FLOAT; // either 4F or 8F
             p_FieldDS = %ADDR(dataStructure)+Qddffibo;
             if Qddffldb = 4;
               json_putDouble(json:%TRIMR(fieldName):fieldDS.Float);
             elseif Qddffldb = 8;
               json_putDouble(json:%TRIMR(fieldName):fieldDS.Double);
             endif;

           when Qddfftyp = FIELD_TYPE_ZONED; // signed (S) type of any supported size
             data = ExtractNumber(%ADDR(dataStructure)+Qddffibo:
                                   'L':'*ZONED':Qddffldd:Qddffldp);
             if qddffldp = 0;
               // If no decimal places, treat like an integer
               if qddffldd > 9;
                 nbrLong = %INT(%TRIMR(data));
                 json_putLong(json:%TRIMR(fieldName):nbrLong);
               else;
                 nbrInt = %INT(%TRIMR(data));
                 json_putInt(json:%TRIMR(fieldName):nbrInt);
               endif;
             else;
               // If decimal places, treat like a float
               nbrDouble = %FLOAT(%TRIMR(data));
               json_putDouble(json:%TRIMR(fieldName):nbrDouble);
             endif;

           when Qddfftyp = FIELD_TYPE_PACKED; // packed (P) type of any supported size
             data = ExtractNumber(%ADDR(dataStructure)+Qddffibo:
                                   'L':'*PACKED':Qddffldd:Qddffldp);
             if qddffldp = 0;
               // If no decimal places, treat like an integer
               if qddffldd > 9;
                 nbrLong = %INT(%TRIMR(data));
                 json_putLong(json:%TRIMR(fieldName):nbrLong);
               else;
                 nbrInt = %INT(%TRIMR(data));
                 json_putInt(json:%TRIMR(fieldName):nbrInt);
               endif;
             else;
               // If decimal places, treat like a float
               nbrDouble = %FLOAT(%TRIMR(data));
               json_putDouble(json:%TRIMR(fieldName):nbrDouble);
             endif;

           when Qddfftyp = FIELD_TYPE_CHAR; // character (A) of any supported size
             json_putString(json:%TRIMR(fieldName):
                            %TRIMR(%SUBST(dataStructure:Qddffibo+1:Qddffldb)));

           when Qddfftyp = FIELD_TYPE_VARCHAR; // varying char (A) of any supported size
             // Assumes 2 byte length at from of varying field
             // I know IBM added longer fields that can have 4 byte length
             // I don't know how to tell the difference though
             p_varCharLen = %ADDR(dataStructure)+Qddffibo;
             json_putString(json:%TRIMR(fieldName):
                           %TRIMR(%SUBST(dataStructure:Qddffibo+3:varCharLen)));

           when Qddfftyp = FIELD_TYPE_DATE; // date (L) type
             p_FieldDS = %ADDR(dataStructure)+Qddffibo;
             json_util_putDate(json:%TRIMR(fieldName):fieldDS.Date);

           when Qddfftyp = FIELD_TYPE_TIME; // time (T) type
             p_FieldDS = %ADDR(dataStructure)+Qddffibo;
             json_util_putTime(json:%TRIMR(fieldName):fieldDS.Time);

           when Qddfftyp = FIELD_TYPE_TIMESTAMP; // timestamp (Z) type
             p_FieldDS = %ADDR(dataStructure)+Qddffibo;
             json_util_putTimestamp(json:%TRIMR(fieldName):fieldDS.Timestamp);

           other;
               // Unknown type?!?


         endsl;

         // Point to next Field Header
         p_Qddffld = p_Qddffld + Qddfdefl;

       endfor;

       // Free up space
       dealloc p_Buffer;

       return json;
      /end-free
     P                 E


     /**
      * \brief Convert record data to an edited number.
      *
      *  This makes it easy to extract zoned and packed fields from
      *  a substring of a record.
      *
      * \param Field pointer
      * \param Edit code
      * \param Field type (*ZONED or *PACKED)
      * \param Number of digits
      * \param Number of decimal places
      *
      * \return String containing edited value
      *
      * \author Brian Garland
      * 
      * \throws CPF9898 File not found
      * \throws CPF9898 Error building JSON object
      */
     P extractNumber   B
     D                 PI           256A
     D  fieldPtr                       *   value
     D  editCode                      1A   value
     D  fieldType                    10A   value
     D  numDigits                     2P 0 value
     D  numDecimals                   2P 0 value

     D editMask        S            256A
     D editMaskLen     S             10I 0
     D outputLen       S             10I 0
     D zeroFillChar    S              1A
     D returnValue     S            256A
     D inputField      S             31A   BASED(fieldPtr)

      /free

       clear returnValue;

       // Invoke the convert edit code API
       clear QUSEC;
       QUSBPRV = %SIZE(QUSEC);
       ConvertEditCode(editMask :
                       editMaskLen :
                       outputLen :
                       zeroFillChar:
                       editCode :
                       ' ' :
                       numDigits :
                       numDecimals :
                       QUSEC);

       if QUSBAVL > 0;
         return *blanks;
       endif;

       // Invoke the Edit API
       clear QUSEC;
       QUSBPRV = %SIZE(QUSEC);
       Edit(returnValue :
            outputLen :
            inputField :
            fieldType :
            numDigits :
            editMask :
            editMaskLen :
            zeroFillChar :
            QUSEC);

       if QUSBAVL > 0;
         return *blanks;
       endif;

       return returnValue;
      /end-free
     P                 E
     